import numpy as np
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
import numpy.ma as ma
from numpy.random import uniform, seed
import os
from matplotlib.colors import hsv_to_rgb
import colorsys
from skimage import measure
import matplotlib.ticker as ticker
import matplotlib.font_manager as fm
from colour import Color
import math
import shutil
import matplotlib.patches as patches
from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
import matplotlib.patches as patches

"""
======================
unicode symbols:
======================
"""
alpha_uni = '$\u03B1$'

"""
======================
Definitions
======================
"""
def cm2inch(value):
    return value/2.54

Aalto_yellow = '#FECB00'
Aalto_orange = '#FF7900'
Aalto_red = '#ED2939'
Aalto_violet = '#B1059D'
Aalto_purple = '#6639B7'
Aalto_blue = '#0065BD'
Aalto_turqoise = '#00A8B4'
Aalto_green = '#009B3A'
Aalto_light_green = '#69BE28'
Aalto_grey = '#928B81'

"""
======================
1. input file
======================
"""
file = 'DW_data0-2.txt'
headerrows = 2

Ko_1p8e6 = 'omfs_32nm/Ko_1p8e6_w_32.omf'
Ko_1p9e6 = 'omfs_32nm/Ko_1p9e6_w_32.omf'
Ko_2p0e6 = 'omfs_32nm/Ko_2p0e6_w_32.omf'

"""
2. define plotting parameters:
"""

#size of square plot:
x_pixels = 1000
y_pixels = 900
#set a phase to rotate the colourwheel in degrees
phase = 180
#set saturation and value of the colours
saturation = 0.6
value = 1
#choose font
# parameters related to the data
para1_min = 0.0
para1_max = 5e6
para1_tick_number = 5
para1_exponent = 6
para1_label = '$\mathit{K_{\mathrm{pp}}}$ [x 10$^'+ str(para1_exponent) +'$ J/m$^{3}$]'
para1_altlabel = '$\mathit{K_{\mathrm{pp,eff}}}$ [x 10$^'+ str(para1_exponent) +'$ J/m$^{3}$]'
para1_altshift = 628318.53
para2_min = 10
para2_max = 200
para2_tick_number = 5
para2_exponent = 0
para2_label = '$\mathit{w}$ [nm]'

label_fontsize = 8
tick_fontsize = 6

contourlinewidth = 0.5
plotlinewidth = 1

#rescaling of OOP contrast: 1 is white, 0 is black.
newwhite = 0.9  
newblack = 0.4
#rescaling of OOP importance. The higher the number, the narrower the DW looks.
OOP_exponent = 1

#parameters for showing arrows in square grid
arrowsampling = 5
arrowoffsetx = 0
arrowoffsety = 0
arrowlength = 4
arrowheadwidth = 1.5
arrowheadlength = 2
arrowwidth = 0.3
OOPthreshold = 80       #show IP arrows only for Mz below threshold (in percent)



# define panel size and position:
width = 7.7
heightratio = 1.20
figuresize=(cm2inch(width), cm2inch(heightratio*width))
figure2size = (cm2inch(width), cm2inch(width))
panel_a = [0.0, 0.12, 1, 1]
panel_b = [0.00, -0.03, 0.32, 0.24]
panel_c = [0.34, -0.03, 0.32, 0.24]
panel_d = [0.68, -0.03, 0.32, 0.24]
xlabelpad = -1
ylabelpad = 0


"""
2. create folders
"""


d = np.loadtxt(file,skiprows=headerrows)

x = d[:,0]
y = d[:,1]
z = d[:,7]
# define grid.
xi = np.linspace(para1_min,para1_max,x_pixels)
yi = np.linspace(para2_min,para2_max,y_pixels)
# grid the data.
zi = griddata((x, y), z, (xi[None,:], yi[:,None]), method='nearest')

# convert to HSV
imageHue = ((zi+180)/360+(phase/360))%1
imageSat = np.full_like(zi, saturation)
imageVal = np.full_like(zi, value)
for m in range(0,imageHue.shape[0]):
    for n in range(0,imageHue.shape[1]):
        if zi[m,n] > 390 :
            imageSat[m,n] = 0
            imageVal[m,n] = 1
#combine HSV and convert to RGB
imageHSV = np.stack((imageHue,imageSat,imageVal),axis=2)
imageRGB = hsv_to_rgb(imageHSV)

"""
3. image processing function:
"""
def plotOMF(filename):
    #set a phase to rotate the colourwheel in degrees
    phase = 180
    #rescaling of OOP contrast: 1 is white, 0 is black.
    newwhite = 0.9  
    newblack = 0.4
    #rescaling of OOP importance. The higher the number, the narrower the DW looks.
    OOP_exponent = 1
    xnodes = 200
    ynodes = 16
    znodes = 1
    xstepsize = 2e-9
    ystepsize = 2e-9
    zstepsize = 1e-9
    #centre of image in multiples of the stepsize. 
    xcentre = int(xnodes/4)
    ycentre = int(ynodes/2)
    zcentre = int(znodes/2)
    #extent of image from centre in multiples of the stepsize. The one that is perpendicular to image needs to be zero:
    xsize = int(xnodes/4)
    ysize = int(ynodes/2*0.8)
    zsize = 0
    #parameters for showing arrows in custom positions
    arrowpositions_init = np.array([[50,10],[50,30],[50,50],[50,70]])

    #import data
    txtname = filename[:-3] + 'txt'
    shutil.copy(filename,txtname)
    d = np.loadtxt(txtname)
    os.remove(txtname)
    dd = np.empty((znodes,ynodes,xnodes,4))
    for zz in range(0,znodes):
        for yy in range(0,ynodes):
            for xx in range(0,xnodes):
                Mx = d[xx+yy*xnodes+zz*xnodes*ynodes,0]
                My = d[xx+yy*xnodes+zz*xnodes*ynodes,1]
                Mz = d[xx+yy*xnodes+zz*xnodes*ynodes,2]
                Ms = math.sqrt(Mx**2+My**2+Mz**2)
                mx = np.clip(Mx/Ms,-1,1)
                my = np.clip(My/Ms,-1,1)
                mz = np.clip(Mz/Ms,-1,1)
                IP_angle = (np.clip(((math.atan2(my,mx)*180/math.pi))+180,0,360)+phase)%360
                dd[zz,yy,xx,0] = mx
                dd[zz,yy,xx,1] = my
                dd[zz,yy,xx,2] = np.sign(mz)*(abs(mz))**(1/OOP_exponent)
                dd[zz,yy,xx,3] = IP_angle
    #extract the part of the array to plot.
    ddd = dd[zcentre,int(ycentre)-int(ysize):int(ycentre)+int(ysize),int(xcentre)-int(xsize):int(xcentre)+int(xsize),:]
    # convert to HSV for given slice
    imageHue = ddd[:,:,3]/360
    imageSat = (1-ddd[:,:,2])/2
    imageVal = (ddd[:,:,2]+1)/2
    # scale colors to follow circlular arc
    imageSat = np.sqrt(1-(1-imageSat)**2)
    imageVal = np.sqrt(1-(1-imageVal)**2)
    #combine HSV and convert to RGB
    imageHSV = np.stack((imageHue,imageSat,imageVal),axis=2)
    imageRGB = hsv_to_rgb(imageHSV)
    #rescale OOP contrast
    for n in range(0,imageRGB.shape[0]):
        for m in range(0,imageRGB.shape[1]):
            temp_YIQ = colorsys.rgb_to_yiq(imageRGB[n,m,0],imageRGB[n,m,1],imageRGB[n,m,2])
            temp_YIQ = (temp_YIQ[0]*(newwhite-newblack)+newblack,temp_YIQ[1],temp_YIQ[2])
            imageRGB[n,m] = colorsys.yiq_to_rgb(temp_YIQ[0],temp_YIQ[1],temp_YIQ[2])
    #add magnetisation components to arrowlist and recentre
    arrowpositions = arrowpositions_init
    mcomponents = np.zeros((arrowpositions.shape[0],3))
    for n in range(0,arrowpositions.shape[0]):
        astart = int(arrowpositions[n,0]-arrowsampling/2)
        aend = int(arrowpositions[n,0]+arrowsampling/2)
        bstart = int(arrowpositions[n,1]-arrowsampling/2)
        bend = int(arrowpositions[n,1]+arrowsampling/2)
        #print(astart,aend)
        mcomponents[n,0] = np.average(ddd[bstart:bend, astart:aend, 0])*(arrowlength)
        mcomponents[n,1] = np.average(ddd[bstart:bend, astart:aend, 1])*(arrowlength)
        mcomponents[n,2] = np.average(ddd[bstart:bend, astart:aend, 2])
        shift1 = mcomponents[n,0]*(1+arrowheadlength/arrowlength)/2
        shift2 = mcomponents[n,1]*(1+arrowheadlength/arrowlength)/2
        #arrowpositions[n,0] = arrowpositions[n,0]- mcomponents[n,0]/2
        #arrowpositions[n,1] = arrowpositions[n,1]- mcomponents[n,1]/2
        #arrowpositions[n,0] = arrowpositions[n,0]- shift1
        #arrowpositions[n,1] = arrowpositions[n,1]- shift2
        
    arrowpositions = np.column_stack((arrowpositions, mcomponents))

    return imageRGB , arrowpositions

"""
create image:
"""

fig = plt.figure(num=None, figsize=figuresize, dpi=80, facecolor='w', edgecolor='k')
"""
phase diagram
"""

ax1 = fig.add_axes(panel_a, label ='1')
ax1.imshow(imageRGB,origin = 'lower')
#ax11 = ax1.twiny()
#relabel ticks of bottom x-axis:
ax1.set_xlim(0, x_pixels)
ax1.xaxis.set_major_locator(plt.MultipleLocator(x_pixels/para1_tick_number))
x_ticks = ticker.FuncFormatter(lambda x, pos: '{0:g}'.format(x/x_pixels*para1_max/(10**para1_exponent)))
ax1.xaxis.set_major_formatter(x_ticks)
#relabel ticks of y-axis
ax1.set_ylim(0, y_pixels)
'''
ax1.yaxis.set_major_locator(plt.MultipleLocator(y_pixels/para2_tick_number))
y_ticks = ticker.FuncFormatter(lambda y, pos: '{0:g}'.format(y/y_pixels*para2_max))
ax1.yaxis.set_major_formatter(y_ticks)
'''
a = y_pixels/(para2_max-para2_min)
b = -para2_min*a
ax1.set_yticks([b+a*10,b+a*50,b+a*100,b+a*150,b+a*200])
ax1.set_yticklabels(['10','50','100','150','200'])

# set labels of axis
ax1.set_xlabel(para1_label, fontsize = label_fontsize,labelpad=xlabelpad)
ax1.set_ylabel(para2_label, fontsize = label_fontsize,labelpad=ylabelpad)
#ax1.yaxis.set_label_position('right')
# position ticks
ax1.tick_params(bottom=True, top=True, left=True, right=True, labelbottom=True, labeltop=False, labelleft=True, labelright=False, direction='in', labelsize=tick_fontsize)
# add text labels, this needs to be done manually!
plt.text(x_pixels*0.3,y_pixels*0.6,'Bloch',fontsize=label_fontsize)
plt.text(x_pixels*0.17,y_pixels*0.08,'Néel',fontsize=label_fontsize)
plt.text(x_pixels*0.04,y_pixels*0.5,'IP',fontsize=label_fontsize)

"""
add inset
"""
axins = zoomed_inset_axes(ax1, 3, loc=5) # zoom = 6
axins.imshow(imageRGB, interpolation="nearest", origin="lower")

# sub region of the original image in pixels
x1, x2, y1, y2 = 350, 450, 60, 160
axins.set_xlim(x1, x2)
axins.set_ylim(y1, y2)

axins.xaxis.set_major_locator(plt.MultipleLocator(x_pixels/para1_tick_number))
axins.xaxis.set_major_formatter(x_ticks)

axins.set_yticks([b+a*50])
axins.set_yticklabels(['50'])

axins.tick_params(bottom=False, top=False, left=False, right=False, labelbottom=False, labeltop=False, labelleft=False, labelright=False, direction='in', labelsize=tick_fontsize)

# draw a bbox of the region of the inset axes in the parent axes and
# connecting lines between the bbox and the inset axes area
mark_inset(ax1, axins, loc1=2, loc2=4, fc="none", ec="0.7")


"""
add simulation files
"""

Ko_1p8e6 = 'omfs_32nm/Ko_1p8e6_w_32.omf'
Ko_1p9e6 = 'omfs_32nm/Ko_1p9e6_w_32.omf'
Ko_2p0e6 = 'omfs_32nm/Ko_2p0e6_w_32.omf'

"""
define plotting parameters:
"""
#size of plot in nm:
x_nm = 100.
y_nm = 32.
z_nm = 0.
#bottom left of plot in nm:
x_left = 50.
y_bottom = 0.
z_bottom = 0.



"""
3. image processing function:
"""
def plotOMF(filename):
    #set a phase to rotate the colourwheel in degrees
    phase = 180
    #rescaling of OOP contrast: 1 is white, 0 is black.
    newwhite = 0.9  
    newblack = 0.4
    #rescaling of OOP importance. The higher the number, the narrower the DW looks.
    OOP_exponent = 1
    xnodes = 200
    ynodes = 16
    znodes = 1
    xstepsize = 2e-9
    ystepsize = 2e-9
    zstepsize = 1e-9
    #centre of image in multiples of the stepsize. 
    xcentre = int(xnodes/4)
    ycentre = int(ynodes/2)
    zcentre = int(znodes/2)
    #extent of image from centre in multiples of the stepsize. The one that is perpendicular to image needs to be zero:
    xsize = int(xnodes/8)
    ysize = int(ynodes/2)
    zsize = 0
    #parameters for showing arrows in custom positions
    arrowpositions_init = np.array([[25,4],[25,11]])

    #import data
    txtname = filename[:-3] + 'txt'
    shutil.copy(filename,txtname)
    d = np.loadtxt(txtname)
    os.remove(txtname)
    dd = np.empty((znodes,ynodes,xnodes,4))
    for zz in range(0,znodes):
        for yy in range(0,ynodes):
            for xx in range(0,xnodes):
                Mx = d[xx+yy*xnodes+zz*xnodes*ynodes,0]
                My = d[xx+yy*xnodes+zz*xnodes*ynodes,1]
                Mz = d[xx+yy*xnodes+zz*xnodes*ynodes,2]
                Ms = math.sqrt(Mx**2+My**2+Mz**2)
                mx = np.clip(Mx/Ms,-1,1)
                my = np.clip(My/Ms,-1,1)
                mz = np.clip(Mz/Ms,-1,1)
                IP_angle = (np.clip(((math.atan2(my,mx)*180/math.pi))+180,0,360)+phase)%360
                dd[zz,yy,xx,0] = mx
                dd[zz,yy,xx,1] = my
                dd[zz,yy,xx,2] = np.sign(mz)*(abs(mz))**(1/OOP_exponent)
                dd[zz,yy,xx,3] = IP_angle
    #extract the part of the array to plot.
    ddd = dd[zcentre,int(ycentre)-int(ysize):int(ycentre)+int(ysize),int(xcentre)-int(xsize):int(xcentre)+int(xsize),:]
    # convert to HSV for given slice
    imageHue = ddd[:,:,3]/360
    imageSat = (1-ddd[:,:,2])/2
    imageVal = (ddd[:,:,2]+1)/2
    # scale colors to follow circlular arc
    imageSat = np.sqrt(1-(1-imageSat)**2)
    imageVal = np.sqrt(1-(1-imageVal)**2)
    #combine HSV and convert to RGB
    imageHSV = np.stack((imageHue,imageSat,imageVal),axis=2)
    imageRGB = hsv_to_rgb(imageHSV)
    #rescale OOP contrast
    for n in range(0,imageRGB.shape[0]):
        for m in range(0,imageRGB.shape[1]):
            temp_YIQ = colorsys.rgb_to_yiq(imageRGB[n,m,0],imageRGB[n,m,1],imageRGB[n,m,2])
            temp_YIQ = (temp_YIQ[0]*(newwhite-newblack)+newblack,temp_YIQ[1],temp_YIQ[2])
            imageRGB[n,m] = colorsys.yiq_to_rgb(temp_YIQ[0],temp_YIQ[1],temp_YIQ[2])
    #add magnetisation components to arrowlist and recentre
    arrowpositions = arrowpositions_init
    mcomponents = np.zeros((arrowpositions.shape[0],3))
    for n in range(0,arrowpositions.shape[0]):
        astart = int(arrowpositions[n,0]-arrowsampling/2)
        aend = int(arrowpositions[n,0]+arrowsampling/2)
        bstart = int(arrowpositions[n,1]-arrowsampling/2)
        bend = int(arrowpositions[n,1]+arrowsampling/2)
        #print(astart,aend)
        mcomponents[n,0] = np.average(ddd[bstart:bend, astart:aend, 0])*(arrowlength)
        mcomponents[n,1] = np.average(ddd[bstart:bend, astart:aend, 1])*(arrowlength)
        mcomponents[n,2] = np.average(ddd[bstart:bend, astart:aend, 2])
        
    arrowpositions = np.column_stack((arrowpositions, mcomponents))
    return imageRGB , arrowpositions



# Ko_1p8e6
ax2 = fig.add_axes(panel_b, label ='2')
img_Ko_1p8e6 , arrows_Ko_1p8e6 = plotOMF(Ko_1p8e6)
for n in range(0,arrows_Ko_1p8e6.shape[0]):
    ax2.arrow(arrows_Ko_1p8e6[n,0]-3,arrows_Ko_1p8e6[n,1],arrows_Ko_1p8e6[n,2],arrows_Ko_1p8e6[n,3],head_width=arrowheadwidth, head_length=arrowheadlength, width = arrowwidth, color ='black')
ax2.imshow(img_Ko_1p8e6,origin = 'lower')
ax2.axis('off')
ax2.annotate('32\nnm',xy =(-0.10,0.5),xycoords = 'axes fraction', ha='center', va='center', rotation=0, fontsize=tick_fontsize)
scalebar = patches.Rectangle((33,2),12.5,3,linewidth=1,edgecolor = 'none',facecolor='white')
ax2.add_patch(scalebar)
ax2.text(32,7,'25 nm',fontsize=tick_fontsize, color = 'white')
ax2.annotate('$\mathit{K_{\mathrm{pp}}} =$1.8 x 10$^6$ J/m$^{3}$',xy =(0.5,-0.3),xycoords = 'axes fraction', ha='center', va='center', rotation=0, fontsize=tick_fontsize)


# Ko_1p9e6
ax3 = fig.add_axes(panel_c, label ='3')
img_Ko_1p9e6 , arrows_Ko_1p9e6 = plotOMF(Ko_1p9e6)
for n in range(0,arrows_Ko_1p9e6.shape[0]):
    ax3.arrow(arrows_Ko_1p9e6[n,0]-2,arrows_Ko_1p9e6[n,1]-2,arrows_Ko_1p9e6[n,2],arrows_Ko_1p9e6[n,3],head_width=arrowheadwidth, head_length=arrowheadlength, width = arrowwidth, color ='black')
ax3.imshow(img_Ko_1p9e6,origin = 'lower')
ax3.axis('off')
ax3.annotate('1.9 x 10$^6$ J/m$^{3}$',xy =(0.50,-0.3),xycoords = 'axes fraction', ha='center', va='center', rotation=0, fontsize=tick_fontsize)

# Ko_2p0e6
ax4 = fig.add_axes(panel_d, label ='4')
img_Ko_2p0e6 , arrows_Ko_2p0e6 = plotOMF(Ko_2p0e6)
for n in range(0,arrows_Ko_2p0e6.shape[0]):
    ax4.arrow(arrows_Ko_2p0e6[n,0],arrows_Ko_2p0e6[n,1]-3,arrows_Ko_2p0e6[n,2],arrows_Ko_2p0e6[n,3],head_width=arrowheadwidth, head_length=arrowheadlength, width = arrowwidth, color ='black')
ax4.imshow(img_Ko_2p0e6,origin = 'lower')
ax4.axis('off')
ax4.annotate('2.0 x 10$^6$ J/m$^{3}$',xy =(0.50,-0.3),xycoords = 'axes fraction', ha='center', va='center', rotation=0, fontsize=tick_fontsize)



"""
add panel labels
"""
axlabels = fig.add_axes([0,0,1,1],label ='labels')
axlabels.axis('off')
axlabels.patch.set_alpha(0.0)
axlabels.text(-0.1,0.95,'(a)',fontsize=label_fontsize)
axlabels.text(-0.1,0.12,'(b)',fontsize=label_fontsize)
axlabels.tick_params(bottom=False, top=False, left=False, right=False, labelbottom=False, labeltop=False, labelleft=False, labelright=False)




"""
show or save image
"""  
#plt.show()
plt.savefig('Ko-vs-w_phase_'+str(phase)+'.pdf', bbox_inches = 'tight',pad_inches = 0)
plt.clf()
